(**********************************************************)
(*							  *)
(*     JPI Modula - 2   ⥪                        *)
(*							  *)
(*	. "", 〈, 㩡襢, . 32-99-79 ()	  *)
(*	  "ଠ⨪", 㩡襢, .32-90-81 () *)
(*							  *)
(*  M  :	Mouse - ᯥ稢 p              *)
(*		p⢮ "".       		  *)
(*               p 室 pp              *)
(*              p MicroSoft  ᮢ⨬  .    *)
(*							  *)
(*     :  1.0 - p  91                            *)
(*							  *)
(*   p : ᮢ ..                          *)
(*							  *)
(*   ᫥ : 22.03.91,  ᮢ ..	  *)
(*							  *)
(**********************************************************)
(******************************************************************************)
(*                                                                            *)
(*       MODULA-2    Library                                                  *)
(*                                                                            *)
(*              LOGITECH SA,   CH-1111 Romanel    (Switzerland)               *)
(*              LOGITECH Inc,  Fremont, CA 94555  (USA)                       *)
(*                                                                            *)
(*       Module  : Mouse                            			      *)
(* 	 The functions implemented in this module provide a		      *)	
(*	 Modula-2 interface for the LOGITECH Mouse Driver.		      *)
(*	 This driver interface is compatible with the Microsoft		      *)
(*	 Mouse Driver interface, so this module can be used with	      *)
(*	 all the compatible mouse drivers.				      *)
(*	 For detailed description of these functions, please		      *)
(*	 refer to your mouse documentation:				      *)
(*									      *)
(*	 e.g. LOGITECH Mouse Driver Programmer's Reference Manual	      *)
(*       Microsoft Mouse, Installation and Operation Manual		      *)
(*									      *)
(*									      *)
(*	 Microsoft is a registered trademark of				      *)
(*	 Microsoft Corporation						      *)
(*									      *)
(*                                                                            *)
(*       Release :    3.0 - July 87                                           *)
(*                                                                            *)
(*       Copyright (C)  1987 Logitech, All rights reserved                    *)
(*                                                                            *)
(*       Permission is hereby granted to registered users to use or abstract  *)
(*       the following program in the implementation of customized versions.  *)
(*       This permission does not include the right to redistribute the       *)
(*       source code of this program.                                         *)
(*                                                                            *)
(******************************************************************************)

IMPLEMENTATION MODULE Mouse;

  FROM SYSTEM IMPORT Registers,Ofs,Seg;

  FROM Lib IMPORT Intr;

  VAR r:Registers;

  PROCEDURE FlagReset(VAR mouseStatus: INTEGER;
                      VAR numberOfButtons: CARDINAL);
  BEGIN
    r.AX:=0;
    Intr(r,33H);
    mouseStatus:=r.AX;
    numberOfButtons := r.BX
  END FlagReset;

  PROCEDURE ShowCursor;
  BEGIN
    r.AX:=1;
    Intr(r,33H)
  END ShowCursor;

  PROCEDURE HideCursor;
  BEGIN
    r.AX:=2;
    Intr(r,33H)
  END HideCursor;

  PROCEDURE GetPosBut(VAR buttonStatus: ButtonSet;
                      VAR horizontal, vertical:INTEGER);
  BEGIN
    r.AX:=3;
    Intr(r,33H);
    buttonStatus:=ButtonSet(r.BX);
    horizontal:=r.CX;
    vertical:=r.DX
  END GetPosBut;

  PROCEDURE SetCursorPos(horizontal, vertical: INTEGER);
  BEGIN
    r.AX:=4;
    r.CX:= horizontal;
    r.DX:= vertical;
    Intr(r,33H)
  END SetCursorPos;

  PROCEDURE GetButPres(button: Button;
                       VAR buttonStatus: ButtonSet;
                       VAR buttonPressCount: CARDINAL;
                       VAR horizontal, vertical: INTEGER);
  BEGIN
    r.BX:= ORD(button);
    r.AX:= 5;
    Intr(r,33H);
    buttonPressCount:=r.BX;
    buttonStatus:=ButtonSet(r.AX);
    horizontal:=r.CX;
    vertical:=r.DX
  END GetButPres;

  PROCEDURE GetButRel(button: Button;
                      VAR buttonStatus: ButtonSet;
                      VAR buttonReleaseCount: CARDINAL;
                      VAR horizontal ,vertical: INTEGER);
  BEGIN
    r.BX:= ORD(button);
    r.AX:= 6;
    Intr(r,33H);
    buttonReleaseCount:=r.BX;
    buttonStatus:=ButtonSet(r.AX);
    horizontal:=r.CX;
    vertical:=r.DX
  END GetButRel;

  PROCEDURE SetHorizontalLimits(minPos, maxPos: INTEGER);
  BEGIN
    r.AX:= 7;
    r.CX:= minPos;
    r.DX:= maxPos;
    Intr(r,33H);
  END SetHorizontalLimits;

  PROCEDURE SetVerticalLimits(minPos, maxPos: INTEGER);
  BEGIN
    r.AX:= 8;
    r.CX:= minPos;
    r.DX:= maxPos;
    Intr(r,33H);
  END SetVerticalLimits;


  PROCEDURE SetGraphicCursor(VAR cursor: GraphicCursor);
    VAR x,y: INTEGER;
  BEGIN
    WITH cursor DO
      x := hotX; y := hotY
    END;
    r.DX:=Ofs(cursor.screenMask);
    r.ES:=Seg(cursor.screenMask);
    r.CX:=y;
    r.BX:=x;
    r.AX:=9;
    Intr(r,33H);
  END SetGraphicCursor;

  PROCEDURE SetTextCursor(selectedCursor,
                          screenMaskORscanStart,
                          cursorMaskORscanStop: CARDINAL);
  BEGIN
    r.AX:= 10;
    r.BX:= selectedCursor;
    r.CX:= screenMaskORscanStart;
    r.DX:= cursorMaskORscanStop;
    Intr(r,33H)
  END SetTextCursor;

  PROCEDURE ReadMotionCounters(VAR horizontal,
                                   vertical:INTEGER);
  BEGIN
    r.AX:= 11;
    Intr(r,33H);
    horizontal:=r.CX;
    vertical:=r.DX;
  END ReadMotionCounters;


  VAR
    PrivateHandler: RECORD
                      PushAX, PushBX, PushCX, PushDX: BYTE;
                      FarCall: BYTE;
                      handlerProc: EventHandler;
                      FarRet: BYTE;
                    END;


  PROCEDURE SetEventHandler(mask: EventSet;
                            handler: EventHandler);
    VAR subr: ADDRESS;
  BEGIN
    PrivateHandler.handlerProc := handler;
    subr := ADR(PrivateHandler);
    r.CX:= CARDINAL(mask);
    r.DX:= Ofs(subr^);
    r.ES:= Seg(subr^);
    r.AX:= 12;
    Intr(r,33H)
  END SetEventHandler;

  PROCEDURE LightPenOn;
  BEGIN
    r.AX:=13;
    Intr(r,33H)
  END LightPenOn;

  PROCEDURE LightPenOff;
  BEGIN
    r.AX:=14;
    Intr(r,33H)
  END LightPenOff;

  PROCEDURE SetMickeysPerPixel(horPix, verPix: CARDINAL);
  BEGIN
    r.AX:= 15;
    r.CX:= horPix;
    r.DX:= verPix;
    Intr(r,33H)
  END SetMickeysPerPixel;

  PROCEDURE ConditionalOff(left, top,
                           right, bottom: INTEGER);
  BEGIN
    r.AX:= 16;
    r.CX:= left;
    r.DX:= top;
    r.SI:= right;
    r.DI:= bottom;
    Intr(r,33H)
  END ConditionalOff;

  PROCEDURE SetSpeedThreshold(threshold: CARDINAL);
  BEGIN
    r.AX:= 19;
    r.DX:= threshold;
    Intr(r,33H)
  END SetSpeedThreshold;


  CONST
    IRET = 0CFH;
  VAR
    MouseVector[0:204]: ADDRESS;
    TerminateProc[0:88H]: PROC;
    OldTermProc: PROC;


  PROCEDURE Terminator;
  BEGIN
    r.AX:= 2;
    Intr(r,33H);      (* hide cursor *)
    r.AX:= 0;
    Intr(r,33H);      (* reset mouse, disable user defined event handler, etc. *)
    TerminateProc := OldTermProc;
    Intr(r,22H)
  END Terminator;

BEGIN
  DriverInstalled := (Ofs(MouseVector^) <> 0) AND
                     (Seg(MouseVector^) <> 0) AND
                     (CARDINAL(MouseVector^) MOD 256 <> IRET);
  IF DriverInstalled THEN
    OldTermProc := TerminateProc;
    TerminateProc := Terminator;
    WITH PrivateHandler DO
      PushAX := BYTE(50H);
      PushBX := BYTE(53H);
      PushCX := BYTE(51H);
      PushDX := BYTE(52H);
      FarCall := BYTE(9AH);
      FarRet := BYTE(0CBH)
    END (* WITH *)
  END
END Mouse.
